/*-
 * Copyright (c) 2015 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Author: Julian Wiesener <jw@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <pthread.h>

#include "vtstor.h"

#include "vsha256.h"

struct vtstor_node {
	unsigned magic;
#define VMOD_VTSTOR_NODE_MAGIC 0x8625cfc7
	char* key;
	char* value;
	size_t key_capacity;
	size_t value_capacity;
	time_t ctime;
	struct vtstor_node *next;
	pthread_rwlock_t lock;
};

struct vtstor_map {
	struct vtstor_node *node;
};

static struct vtstor_node* create_node(struct vmod_vtstor_vtstor *vmi)
{
	(void) vmi;
	struct vtstor_node* node = malloc(sizeof(struct vtstor_node));
	if(!node)
		return NULL;
	memset(node, 0, sizeof(struct vtstor_node));
	node->magic = 0x8625cfc7;
	pthread_rwlock_init(&node->lock, NULL);
	return node;
}

static struct vtstor_node* delete_node_r(struct vmod_vtstor_vtstor *vmi, struct vtstor_node *node)
{
	CHECK_OBJ_NOTNULL(node, VMOD_VTSTOR_NODE_MAGIC);
	if(node->next)
		node->next = delete_node_r(vmi, node->next);
	if(node->key)
		vmi->dealloc(vmi, &node->key);
	if(node->value)
		vmi->dealloc(vmi, &node->value);
	pthread_rwlock_destroy(&node->lock);
	free(node);
	return 0;
}

static inline struct vtstor_node* find_bucket(struct vmod_vtstor_vtstor *vmi, uint32_t hash)
{
	return (vmi->map[hash % vmi->buckets].node);
}

/* By: Nils Goroll <slink@uplex.de> */
#define SMU_PWR_MIN     6      /*  64 b */
#define SMU_PWR_MAX    21      /*   2 M */
static inline int 
szpwr(size_t size, size_t *space) {
	int pwr;

	if(size < (1 << SMU_PWR_MIN)) {
		pwr = SMU_PWR_MIN;
		*space = (1 << SMU_PWR_MIN);
	} else if (size > (1 << SMU_PWR_MAX)) {
		return -1;
	} else {
		size_t s;
		pwr   = log2(size);
		s     = (1 << pwr);
		if((s < size) && (pwr < SMU_PWR_MAX)) {
			pwr++; 
			*space = (s << 1);
		} else {
		*space = s;
		}
	}

	return pwr;
}

static struct vtstor_node* find_node(struct vtstor_node *node, VCL_STRING key)
{
	CHECK_OBJ_NOTNULL(node, VMOD_VTSTOR_NODE_MAGIC);
	AN(key);
	pthread_rwlock_rdlock(&node->lock);
	if(node->key && !strcmp(node->key, key))
		return node;
	pthread_rwlock_unlock(&node->lock);
	if(node->next == NULL)
		return NULL;
	return find_node(node->next, key);
}

static struct vtstor_node* new_node(struct vmod_vtstor_vtstor *vmi, struct vtstor_node *node)
{
	CHECK_OBJ_NOTNULL(node, VMOD_VTSTOR_NODE_MAGIC);
	time_t now = time(NULL);

	pthread_rwlock_wrlock(&node->lock);
	if(node->ctime + vmi->timeout < now)
		return node;
	
	if(node->next == NULL) {
		struct vtstor_node* newnode = create_node(vmi);
		if(!newnode) {
			pthread_rwlock_unlock(&node->lock);
			WRONG("Could not allocate new node");
		}
		pthread_rwlock_wrlock(&newnode->lock);
		node->next = newnode;
		pthread_rwlock_unlock(&node->lock);
		return node->next;
	}

	pthread_rwlock_unlock(&node->lock);
	return new_node(vmi, node->next);
}

static int vtstor_set_key(struct vmod_vtstor_vtstor *vmi, struct vtstor_node *node, VCL_STRING s)
{
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);
	CHECK_OBJ_NOTNULL(node, VMOD_VTSTOR_NODE_MAGIC);
	size_t size = strlen(s) + 1;
	if(size > vmi->key_max)
		return 0;
	if(node->key) {
		if(node->key_capacity >= size) {
			strcpy(node->key, s);
			return node->key_capacity;
		}
		vmi->dealloc(vmi, &node->key);
	}
	node->key_capacity = vmi->alloc(vmi, size, &node->key);
	if(node->key_capacity >= size)
		strcpy(node->key, s);
	return node->key_capacity;
}

static int vtstor_set_value(struct vmod_vtstor_vtstor *vmi, struct vtstor_node *node, VCL_STRING s)
{
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);
	CHECK_OBJ_NOTNULL(node, VMOD_VTSTOR_NODE_MAGIC);
	size_t size = strlen(s) + 1;
	if(size > vmi->value_max)
		return 0;
	if(node->value) {
		if(node->value_capacity >= size) {
			strcpy(node->value, s);
			return node->value_capacity;
		}
		vmi->dealloc(vmi, &node->value);
	}
	node->value_capacity = vmi->alloc(vmi, size, &node->value);
	if(node->value_capacity >= size)
		strcpy(node->value, s);
	return node->value_capacity;
}

uint32_t vtstor_hash_sha256(VCL_STRING s)
{
	SHA256_CTX ctx;
	unsigned char sha[SHA256_LEN];
	uint32_t hash;

	SHA256_Init(&ctx);
	SHA256_Update(&ctx, (const void *) s, strlen(s));
	SHA256_Final(sha, &ctx);
	memcpy(&hash, sha, 4);

       return (hash ? hash : 1);
}

size_t vtstor_alloc(struct vmod_vtstor_vtstor *vmi, size_t size, char** ptrp)
{
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);
	AN(ptrp);
	AZ(*ptrp);

	size_t space;
	if(szpwr(size, &space) < 1)
		return 0;
	
	char *ptr = malloc(space);
	if(ptr) {
		*ptrp = ptr;
		return space;
	}
	return 0;
}

void vtstor_dealloc(struct vmod_vtstor_vtstor *vmi, char **ptr)
{
	(void) vmi;
	if(*ptr) {
		free(*ptr);
		*ptr = NULL;
	}
}

struct vtstor_map* vtstor_init(struct vmod_vtstor_vtstor *vmi)
{
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);
	struct vtstor_map* map = malloc(sizeof(struct vtstor_map) * vmi->buckets);
	if(!map)
		WRONG("Could not allocate map");

	unsigned u;
	for(u = 0; u < vmi->buckets; u++) {
		map[u].node = create_node(vmi);
		if(!map[u].node)
			WRONG("Could not allocate node");
	}
	return map;
}

void vtstor_deleteall(struct vmod_vtstor_vtstor *vmi)
{
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);
	AN(vmi->map);

	unsigned u;
	for(u = 0; u < vmi->buckets; u++)
		vmi->map[u].node = delete_node_r(vmi, vmi->map[u].node);
	free(vmi->map);
	vmi->map = NULL;
}

void vtstor_store(struct vmod_vtstor_vtstor *vmi, VCL_STRING key, VCL_STRING value)
{
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);
	AN(vmi->map);
	struct vtstor_node *headnode = find_bucket(vmi, vmi->hash(key));
	AN(headnode); 
	struct vtstor_node *node = find_node(headnode, key);
	if(node) {
		CHECK_OBJ_NOTNULL(node, VMOD_VTSTOR_NODE_MAGIC);
		pthread_rwlock_unlock(&node->lock);
		pthread_rwlock_wrlock(&node->lock);

		/* need recheck key, node might got recycled */
		if(node->key && !strcmp(node->key, key)) {
			if(vtstor_set_value(vmi, node, value)) {
				node->ctime = time(NULL);
				pthread_rwlock_unlock(&node->lock);
				return;
			}
			pthread_rwlock_unlock(&node->lock);
			WRONG("could not store value");
		}
		pthread_rwlock_unlock(&node->lock);
	}
	/* if someone had created a node for that key in the meantime, the
	 * keys are not uniq, that should not happen, i don't care for now */
	node = new_node(vmi, headnode);
	CHECK_OBJ_NOTNULL(node, VMOD_VTSTOR_NODE_MAGIC);
	if(!vtstor_set_key(vmi, node, key)) {
		pthread_rwlock_unlock(&node->lock);
		WRONG("could not store key");
	}
	if(!vtstor_set_value(vmi, node, value)) {
		pthread_rwlock_unlock(&node->lock);
		node->key[0] = '\0';
		WRONG("could not store value");
	}
	node->ctime = time(NULL);
	pthread_rwlock_unlock(&node->lock);
}

VCL_STRING vtstor_get(struct vmod_vtstor_vtstor *vmi, VCL_STRING key)
{
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);
	AN(vmi->map);
	struct vtstor_node *node = find_bucket(vmi, vmi->hash(key));
	CHECK_OBJ_NOTNULL(node, VMOD_VTSTOR_NODE_MAGIC);
	node = find_node(node, key);
	if(node == NULL)
		return NULL;

	CHECK_OBJ_NOTNULL(node, VMOD_VTSTOR_NODE_MAGIC);
	pthread_rwlock_unlock(&node->lock);
	/* returning a pointer here is racy, would need to copy into priv */
	return(node->value);
}

void vtstor_delete(struct vmod_vtstor_vtstor *vmi, VCL_STRING key)
{
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);
	AN(vmi->map);
	struct vtstor_node *node = find_bucket(vmi, vmi->hash(key));
	CHECK_OBJ_NOTNULL(node, VMOD_VTSTOR_NODE_MAGIC);
	node = find_node(node, key);
	if(node == NULL)
		return;

	CHECK_OBJ_NOTNULL(node, VMOD_VTSTOR_NODE_MAGIC);
	pthread_rwlock_unlock(&node->lock);
	pthread_rwlock_wrlock(&node->lock);

	/* need recheck key, node might got recycled */
	if(node->key && !strcmp(node->key, key)) {
		node->key[0] = '\0';
		node->ctime = 0;
	}
	pthread_rwlock_unlock(&node->lock);
	return;
}
